#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Provider model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instrument import Instrument


class Provider(BaseModel):
    """ Provider model. """
    instrument = models.ForeignKey(Instrument, null=True,
                                   on_delete=models.SET_NULL)
    name = models.CharField(max_length=250, null=False, blank=False,
                            help_text=_("Provider name."))

    description = models.TextField(null=True, blank=True, default=None,
                                   help_text=_("Provider description."))

    url = models.CharField(max_length=250, null=True, blank=False, default=None,
                           help_text=_("Provider url."))

    version = models.CharField(max_length=15, null=False, blank=False,
                               help_text=_("Provider version."))

    acronym = models.CharField(max_length=30, null=False, blank=False, unique=True,
                               help_text=_("Provider Acronym."))

    queue = models.CharField(max_length=50, null=False, blank=False, unique=True,
                             help_text=_("Queue where provider listens for requests."))

    enabled = models.BooleanField(null=False, blank=False, default=False,
                                  help_text=_("Whether this provider is enabled to be used"))

    allow_validation = models.BooleanField(null=False, blank=False, default=False,
                                           help_text="Whether this provider provides validation feature for data")

    validation_active = models.BooleanField(null=False, blank=False, default=False,
                                            help_text="Whether this provider will be used for validation")

    alert_below = models.DecimalField(null=True, blank=False, decimal_places=2, max_digits=4, default=None,
                                      help_text=_("Mark as alert for result values below this threshold."))

    warning_below = models.DecimalField(null=True, blank=False, decimal_places=2, max_digits=4, default=None,
                                        help_text=_("Mark as warning for result values below this threshold."))

    inverted_polarity = models.BooleanField(null=False, default=False,
                                            help_text=_('If enabled, good values are lower values'))

    image = models.CharField(max_length=250, null=False, blank=False,
                             help_text=_("Provider Docker image."))

    has_service = models.BooleanField(null=False, default=False,
                                      help_text=_('Whether this provider starts a service and must be balanced'))
    service_port = models.IntegerField(null=True, default=None,
                                       help_text=_('Port where service is listening'))

    options_schema = models.TextField(null=True, blank=True, default=None,
                                      help_text=_("Schema for provider options."))
    options = models.TextField(null=True, blank=True, default=None,
                               help_text=_("Provider options."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Provider(name='%s', acronym='%s', enabled='%r', queue='%r', allow_validation='%r')>" % (
            self.name, self.acronym, self.enabled, self.queue, self.allow_validation,
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA API permissions module """
from rest_framework import permissions

from tesla_ce.models import InstitutionUser
from tesla_ce.models import User
from tesla_ce.models import AuthenticatedModule
from tesla_ce.models import Course


class GlobalAdminPermission(permissions.BasePermission):
    """
        Global admins have read only permissions
    """
    def has_permission(self, request, view):
        return request.user.is_staff


class GlobalAdminReadOnlyPermission(GlobalAdminPermission):
    """
        Global admins have read only permissions
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)

        return False


class InstitutionMemberPermission(permissions.BasePermission):
    """
        Only members of the institution can access to the view
    """

    def _get_user_institution(self, request):

        if not request.user.is_authenticated:
            return None

        if isinstance(request.user, InstitutionUser):
            return request.user.institution

        if isinstance(request.user, User):
            try:
                return request.user.institutionuser.institution
            except User.institutionuser.RelatedObjectDoesNotExist:
                return None

    def has_permission(self, request, view):

        institution = self._get_user_institution(request)

        if institution is None or not request.path.startswith('/api/{}/institution/{}/'.format(request.version,
                                                                                               institution.id)):
            return False
        return True


class InstitutionMemberReadOnlyPermission(InstitutionMemberPermission):
    """
        Only members of the institution can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionAdminPermission(InstitutionMemberPermission):
    """
        Only admins of the institution can access to the view
    """
    def has_permission(self, request, view):
        if super().has_permission(request, view):
            if isinstance(request.user, InstitutionUser):
                return request.user.inst_admin
            else:
                return request.user.institutionuser.inst_admin

        return False


class InstitutionAdminReadOnlyPermission(InstitutionAdminPermission):
    """
        Only admins of the institution can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionDataAdminPermission(InstitutionMemberPermission):
    """
        Only admins of the institution can access to the view
    """
    def has_permission(self, request, view):
        if super().has_permission(request, view):
            if isinstance(request.user, InstitutionUser):
                return request.user.data_admin
            else:
                return request.user.institutionuser.data_admin

        return False


class InstitutionDataAdminReadOnlyPermission(InstitutionDataAdminPermission):
    """
        Only admins of the institution can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionLegalAdminPermission(InstitutionMemberPermission):
    """
        Only admins of the institution can access to the view
    """
    def has_permission(self, request, view):
        if super().has_permission(request, view):
            if isinstance(request.user, InstitutionUser):
                return request.user.legal_admin
            else:
                return request.user.institutionuser.legal_admin

        return False


class InstitutionLegalAdminReadOnlyPermission(InstitutionLegalAdminPermission):
    """
        Only admins of the institution can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionSENDAdminPermission(InstitutionMemberPermission):
    """
        Only SEND admins of the institution can access to the view
    """
    def has_permission(self, request, view):
        if super().has_permission(request, view):
            if isinstance(request.user, InstitutionUser):
                return request.user.send_admin
            else:
                return request.user.institutionuser.send_admin

        return False


class InstitutionSENDAdminReadOnlyPermission(InstitutionSENDAdminPermission):
    """
        Only SEND admins of the institution can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionLearnerPermission(InstitutionMemberPermission):
    """
        An institution learner accessing his/her information
    """
    def has_permission(self, request, view):
        if super().has_permission(request, view):
            return '/learner/{}/'.format(request.user.id) in request.path
        return False


class InstitutionLearnerReadOnlyPermission(InstitutionLearnerPermission):
    """
        An institution learner accessing his/her information in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionCourseInstructorPermission(InstitutionMemberPermission):
    """
        An institution instructor accessing his/her course information
    """
    def _get_course(self, request):
        """ Get the course that is accessed """
        institution = super()._get_user_institution(request)
        if institution is None:
            return None
        url_pattern = '/api/{}/institution/{}/course/'.format(request.version, institution.id)
        if request.path.startswith(url_pattern):
            course_part = request.path.split(url_pattern)
            if len(course_part) != 2:
                return None
            course_id_str = course_part[1].split('/')[0]
            try:
                course_id = int(course_id_str)
                return Course.objects.filter(vle__institution=institution).get(id=course_id)
            except TypeError:
                return None
            except Course.DoesNotExist:
                return None
        return None

    def has_permission(self, request, view):
        if super().has_permission(request, view):
            course = self._get_course(request)
            if course is None:
                return False
            try:
                return course.instructors.filter(id=request.user.id).count() == 1
            except Exception:
                return False
        return False


class InstitutionCourseInstructorReadOnlyPermission(InstitutionCourseInstructorPermission):
    """
        An institution instructor accessing his/her course information in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class InstitutionCourseLearnerPermission(InstitutionMemberPermission):
    """
        An institution learner accessing his/her course information
    """
    def _get_course(self, request):
        """ Get the course that is accessed """
        institution = super()._get_user_institution(request)
        if institution is None:
            return None
        url_pattern = '/api/{}/institution/{}/course/'.format(request.version, institution.id)
        if request.path.startswith(url_pattern):
            course_part = request.path.split(url_pattern)
            if len(course_part) != 2:
                return None
            course_id_str = course_part[1].split('/')[0]
            try:
                course_id = int(course_id_str)
                return Course.objects.filter(vle__institution=institution).get(id=course_id)
            except TypeError:
                return None
            except Course.DoesNotExist:
                return None
        return None

    def has_permission(self, request, view):
        if super().has_permission(request, view):
            course = self._get_course(request)
            if course is None:
                return False
            try:
                return course.learners.filter(id=request.user.id).count() == 1
            except Exception:
                return False
        return False


class InstitutionCourseLearnerReadOnlyPermission(InstitutionCourseLearnerPermission):
    """
        An institution learner accessing his/her information in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class VLEPermission(permissions.BasePermission):
    """
        Only target VLE can access to the view
    """

    def has_permission(self, request, view):

        if not isinstance(request.user, AuthenticatedModule):
            return False

        vle_id = None
        if request.user.type == 'vle':
            vle_id = request.user.pk

        if vle_id is None or not request.path.startswith('/api/{}/vle/{}/'.format(request.version, vle_id)):
            return False
        return True


class VLEReadOnlyPermission(VLEPermission):
    """
        Only the targeted VLE can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False


class ProviderPermission(permissions.BasePermission):
    """
        Only target Provider can access to the view
    """

    def has_permission(self, request, view):

        if not isinstance(request.user, AuthenticatedModule):
            return False

        provider_id = None
        if request.user.type == 'provider':
            provider_id = request.user.pk

        if provider_id is None or not request.path.startswith(
                '/api/{}/provider/{}/'.format(request.version, provider_id)
        ):
            return False
        return True


class ProviderReadOnlyPermission(ProviderPermission):
    """
        Only the targeted Provider can access to the view in read only mode
    """
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return super().has_permission(request, view)
        return False



#  Copyright (c) 2020 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for institution informed consent administration """
import logging

import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_informed_consent(rest_api_client, user_global_admin, institution_course_test_case):
    institution_user = institution_course_test_case['user'].institutionuser
    institution = institution_course_test_case['institution']
    institution_id = institution.id

    user_global_admin.is_staff = True
    user_global_admin.save()

    # Set global admin user.
    rest_api_client.force_authenticate(user=user_global_admin)

    # Institution privileges
    institution_user.inst_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    logging.info('\nTESTING INSTITUTIONS: Informed consents *********************************************')
    """ ---------------------------------------------------------------------
     LIST INFORMED CONSENTS:
        GET /api/v2/institution/(int:institution_id)/ic/
        Status Codes:
            200 OK – Ok
            404 Not Found – Institution not found
  
     Privileges: Institution privileges
     """
    logging.info('\n1) LIST INFORMED CONSENTS --------------------------------------')
    str_path = '/api/v2/institution/{}/ic/'.format(institution_id)
    str_response = 'RESPONSE institution_id={}:'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consents', str_response, 200)
    # Ensure that no informed consent exists
    assert body['count'] == 0

    # Create a new informed consent
    """ ---------------------------------------------------------------------
    CREATE A NEW INFORMED CONSENT:
    POST /api/v2/institution/(int:institution_id)/ic/
    Status Codes:
        201 Created – Created
        400 Bad Request – Invalid information provided. The response contains the description of the errors.
        404 Not Found – Institution not found

    Privileges: Institution Legal privileges
    """
    logging.info('\n2) CREATE A NEW INFORMED CONSENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # Version format: X.Y.Z (with X, Y and Z integer values
    # Date format: YYYY-MM-DDThh:mm[:ss[.uuuuuu]][+HH:MM|-HH:MM|Z].']
    str_data = {'version': '0.0.2', 'valid_from': '2021-06-01T10:00'}
    new_ic_id = tests.utils.post_rest_api_client(rest_api_client, str_path,
                                                 str_data, 'Create a new Informed Consent',
                                                 'RESPONSE:', 201)

    # List Informed Consents
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consents', str_response, 200)
    # Ensure Informed Consent exists
    n_ic = body['count']
    assert n_ic > 0

    # Status 400: Bad request
    str_data_bad_request = {'version': '1'}
    tests.utils.post_rest_api_client(rest_api_client, str_path,
                                     str_data_bad_request, 'Create a new Informed Consent',
                                     'RESPONSE:', 400)

    # TODO? Read Informed Consent information errors: status 404

    # Create a new informed consent
    """ ---------------------------------------------------------------------
    READ INFORMED CONSENT INFORMATION:
    GET /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/
    Status Codes:
        200 OK – Ok
        404 Not Found – Institution not found
        404 Not Found – Informed consent not found

    Privileges: Institution privileges
    """
    logging.info('\n3) READ INFORMED CONSENT INFORMATION --------------------------------------')
    institution_user.legal_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/ic/{}/'.format(institution_id, new_ic_id)
    str_response = 'RESPONSE Informed Consent id={}:'.format(new_ic_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Informed Consent Information', str_response, 200)

    # TODO? Read Informed Consent information errors: status 404 institution and/or informed consent not found

    # Update informed consent
    """ ---------------------------------------------------------------------
    UPDATE INFORMED CONSENT:
        PUT /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
    Request Headers: Authorization - JWT with Institution Legal privileges
    """
    logging.info('\n4) UPDATE INFORMED CONSENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_data = {'version': '0.1.666', 'valid_from': '2021-06-06T23:00:00Z'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Update Informed Consent', 'RESPONSE: ', 200)

    # TODO? Update Informed Consent errors: status 400 & 404

    # List informed consent documents
    """ ---------------------------------------------------------------------
    LIST INFORMED CONSENT DOCUMENTS:
        GET /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/document/
        Status Codes:
            200 OK – Ok
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
            404 Not Found – Informed consent document not found
    Request Headers: Authorization - JWT with Institution privileges
    """
    logging.info('\n5) LIST INFORMED CONSENT DOCUMENTS--------------------------------------')
    institution_user.legal_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/ic/{}/document/'.format(institution_id, new_ic_id)
    str_response = 'RESPONSE institution_id={}, informed consent id={}:'.format(institution_id, new_ic_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consent Documents', str_response, 200)
    # Ensure that no informed consent document exists
    assert body['count'] == 0

    # Create a new informed consent document
    """ ---------------------------------------------------------------------
    CREATE A NEW INFORMED CONSENT DOCUMENT:
    The document for each language is provided by HTML and with an attached file. .. 
    http:post:: /api/v2/institution/(int:institution_id)/ic/(int:consent_id)/document/
        Status Codes:
            201 Created – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            400 Not Found – Institution not found
            404 Not Found – Informed consent not found
    Request Headers: Authorization - JWT with Institution Legal privileges
    """
    logging.info('\n6) CREATE A NEW INFORMED CONSENT DOCUMENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # 666 TODO HOW TO ADD THE ATTACHED FILE AS THE NEW INFORMED CONSENT DOCUMENT?
    # str_data = {'version': '0.0.2', 'valid_from': '2021-06-01T10:00'}
    language = 'en'
    str_data = {'language': language}
    new_ic_document_id = tests.utils.post_rest_api_client(rest_api_client, str_path,
                                                          str_data, 'Create a new Informed Consent Document',
                                                          'RESPONSE:', 201)

    # aux = 'new_ic_document_id:{}'.format(new_ic_document_id)
    # tests.utils.print_log(aux)
    # List Informed Consents
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consents', str_response, 200)
    # Ensure Informed Consent exists
    n_ic_document = body['count']
    assert n_ic_document > 0

    # Read informed consent document information
    """ ---------------------------------------------------------------------
    READ INFORMED CONSENT DOCUMENT INFORMATION:
        GET /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/document/(str: language)/
        Status Codes:
            200 OK – Ok
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
            404 Not Found – Informed consent document not found
    Request Headers: Authorization - JWT with Institution privileges
    """
    logging.info('\n7) READ INFORMED CONSENT DOCUMENT INFORMATION --------------------------------------')
    institution_user.legal_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # language = 'en'
    str_path = '/api/v2/institution/{}/ic/{}/document/{}/'.format(institution_id, new_ic_id, language)
    str_response = 'RESPONSE Informed Consent id={}, language={}:'.format(new_ic_id, language)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Informed Consent Document Information',
                                    str_response, 200)

    # TODO Update informed consent document
    """ ---------------------------------------------------------------------
    UPDATE INFORMED CONSENT DOCUMENT:
        PUT /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/document/(str: language)/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
            404 Not Found – Informed consent document not found
    Request Headers: Authorization - JWT with Institution Legal privileges
    """
    logging.info('\n8) UPDATE INFORMED CONSENT DOCUMENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    # 666 TODO how to update an informed consent document? adding a new attached file??

    # TODO Read informed consent document information

    # Delete informed consent document
    """ ---------------------------------------------------------------------
    DELETE INFORMED CONSENT DOCUMENT:
        DELETE /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/document/(str: language)/
        Status Codes:
            200 OK – Ok
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
            404 Not Found – Informed consent document not found
    Request Headers: Authorization - JWT with Institution Legal privileges
    """
    logging.info('\n9) DELETE INFORMED CONSENT DOCUMENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/ic/{}/document/{}/'.format(institution_id, new_ic_id, language)
    # 666 check source https://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.7
    # A successful response SHOULD be 200 (OK) if the response includes an entity describing the status,
    # 202 (Accepted) if the action has not yet been enacted, or 204 (No Content) if the action has been enacted
    # but the response does not include an entity.
    # 666 ADD THOSE STATUS TO DOC?
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete Informed Consent Document', "RESPONSE: ", 204)

    # List Informed Consents
    str_path = '/api/v2/institution/{}/ic/{}/document/'.format(institution_id, new_ic_id)
    str_response = 'RESPONSE institution_id={}, informed consent id={}:'.format(institution_id, new_ic_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consent Documents', str_response, 200)
    logging.info(body)
    # Ensure number of Informed Consent Documents has decreased
    assert body['count'] == (n_ic_document - 1)

    # Delete informed consent
    """ ---------------------------------------------------------------------
    DELETE INFORMED CONSENT:
        DELETE /api/v2/institution/(int: institution_id)/ic/(int: consent_id)/
        Status Codes:
            200 OK – Ok
            404 Not Found – Institution not found
            404 Not Found – Informed consent not found
    Request Headers: Authorization - JWT with Institution Legal privileges
    """
    logging.info('\n10) DELETE INFORMED CONSENT --------------------------------------')
    institution_user.legal_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/ic/{}/'.format(institution_id, new_ic_id)
    # 666 check source https://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.7
    # A successful response SHOULD be 200 (OK) if the response includes an entity describing the status,
    # 202 (Accepted) if the action has not yet been enacted, or 204 (No Content) if the action has been enacted
    # but the response does not include an entity.
    # 666 ADD THOSE STATUS TO DOC?
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete Informed Consent', "RESPONSE: ", 204)

    # List Informed Consents
    str_path = '/api/v2/institution/{}/ic/'.format(institution_id)
    str_response = 'RESPONSE institution_id={}:'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Informed Consents', str_response, 200)
    logging.info(body)
    # Ensure number of Informed Consent has decreased
    assert body['count'] == (n_ic - 1)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Module for Authentication API profile """
import pytest
from tests.utils import get_profile


@pytest.mark.django_db
def test_api_authentication_profile_routes(rest_api_client, api_client, institution_course_test_case,
                                           user_global_admin, base_ui_routes):

    # Get global admin token
    global_admin_token = api_client.get_admin_token_pair(user_global_admin, {})
    assert global_admin_token is not None
    assert 'access_token' in global_admin_token
    global_admin_profile_resp = get_profile(rest_api_client, global_admin_token['access_token'])
    assert global_admin_profile_resp.status_code == 200
    assert 'routes' in global_admin_profile_resp.data
    assert '/admin' in global_admin_profile_resp.data['routes']
    assert '/dashboard' in global_admin_profile_resp.data['routes']
    assert '/mycourses' not in global_admin_profile_resp.data['routes']
    assert '/inst_admin' not in global_admin_profile_resp.data['routes']

    # Get institution user token
    inst_user_token = api_client.get_user_token_pair(institution_course_test_case['user'].institutionuser, {})
    assert inst_user_token is not None
    assert 'access_token' in inst_user_token
    inst_user_profile_resp = get_profile(rest_api_client, inst_user_token['access_token'])
    assert inst_user_profile_resp.status_code == 200
    assert 'institution' in inst_user_profile_resp.data
    assert 'routes' in inst_user_profile_resp.data
    assert len(inst_user_profile_resp.data['routes']) == 1

    # Get institution admin token
    admin_user = institution_course_test_case['user'].institutionuser
    admin_user.inst_admin = True
    admin_user.data_admin = True
    admin_user.legal_admin = True
    admin_user.send_admin = True
    admin_user.save()

    inst_admin_user_token = api_client.get_user_token_pair(admin_user, {})
    assert inst_admin_user_token is not None
    assert 'access_token' in inst_admin_user_token
    inst_admin_user_profile_resp = get_profile(rest_api_client, inst_admin_user_token['access_token'])
    assert inst_admin_user_profile_resp.status_code == 200
    assert 'institution' in inst_admin_user_profile_resp.data
    assert 'routes' in inst_admin_user_profile_resp.data
    assert '/admin' not in inst_admin_user_profile_resp.data['routes']
    assert '/dashboard' in inst_admin_user_profile_resp.data['routes']
    assert '/mycourses' not in inst_admin_user_profile_resp.data['routes']
    assert '/inst_admin' in inst_admin_user_profile_resp.data['routes']

    # Get institution learner token
    learner_user = institution_course_test_case['learner'].institutionuser.learner

    inst_learner_user_token = api_client.get_user_token_pair(learner_user, {})
    assert inst_learner_user_token is not None
    assert 'access_token' in inst_learner_user_token
    inst_learner_user_profile_resp = get_profile(rest_api_client, inst_learner_user_token['access_token'])
    assert inst_learner_user_profile_resp.status_code == 200
    assert 'institution' in inst_learner_user_profile_resp.data
    assert 'routes' in inst_learner_user_profile_resp.data
    assert '/admin' not in inst_learner_user_profile_resp.data['routes']
    assert '/dashboard' in inst_learner_user_profile_resp.data['routes']
    assert '/mycourses' in inst_learner_user_profile_resp.data['routes']
    assert '/inst_admin' not in inst_learner_user_profile_resp.data['routes']

    # Get institution instructor token
    instructor_user = institution_course_test_case['instructor'].institutionuser.instructor

    inst_instructor_user_token = api_client.get_user_token_pair(instructor_user, {})
    assert inst_instructor_user_token is not None
    assert 'access_token' in inst_learner_user_token
    inst_instructor_user_profile_resp = get_profile(rest_api_client, inst_instructor_user_token['access_token'])
    assert inst_instructor_user_profile_resp.status_code == 200
    assert 'institution' in inst_instructor_user_profile_resp.data
    assert 'routes' in inst_instructor_user_profile_resp.data
    assert '/admin' not in inst_instructor_user_profile_resp.data['routes']
    assert '/dashboard' in inst_instructor_user_profile_resp.data['routes']
    assert '/mycourses' in inst_instructor_user_profile_resp.data['routes']
    assert '/inst_admin' not in inst_instructor_user_profile_resp.data['routes']



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test fixtures module """
import json
import os
import random
import string

import pytest
from django.conf import settings
from django.utils import timezone

from tesla_ce.client import Client
from tesla_ce.client import ConfigManager
from tesla_ce.models import Provider


def get_random_string(length=10):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for idx in range(length))


@pytest.fixture(scope='session')
def django_db_setup():
    # Disable automatic database configuration
    pass


def create_module_client(client, module):
    assert client is not None
    credentials = client.vault.get_module_credentials(module)
    assert 'role_id' in credentials and 'secret_id' in credentials
    config = ConfigManager(load_config=False)
    config.config.set('VAULT_SSL_VERIFY', client.config.config.get('VAULT_SSL_VERIFY'))
    config.load_vault(vault_url=client.config.config.get('VAULT_URL'),
                      role_id=credentials['role_id'],
                      secret_id=credentials['secret_id'])
    return Client(config=config)


@pytest.mark.django_db
def get_provider_object(api_client, desc_file):
    """
        Register a provider from a description file
        :param api_client: API client object
        :param desc_file: Path to the description file
        :return: Provider object
    """
    with open(desc_file, 'r') as info_file:
        provider_info = json.load(info_file)
    api_client.register_provider(provider_info)

    provider = Provider.objects.get(acronym=provider_info['acronym'])

    return provider


@pytest.fixture
def providers(api_client, db):
    tfr_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..', '..', 'providers', 'fr_tfr.json'))
    tfa_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..', '..', 'providers', 'fa_tfa.json'))
    tpt_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..', '..', 'providers', 'pt_tpt.json'))

    provider_obj = {
        'fr': get_provider_object(api_client, tfr_file),
        'fa': get_provider_object(api_client, tfa_file),
        'plag': get_provider_object(api_client, tpt_file),
    }

    return provider_obj


@pytest.fixture(scope="session", autouse=True)
def tesla_ce_system(django_db_blocker):
    """
        Initialize the TeSLA client.

        :return: TeSLA Client with Administration credentials
    """
    with django_db_blocker.unblock():
        client = Client(enable_management=True)
        report = client.check_configuration()
        print(json.dumps(report))
        assert report['valid']
        client.initialize()

        # Remove Vault Token from environment if present
        if 'VAULT_TOKEN' in os.environ:
            del os.environ['VAULT_TOKEN']

        yield client


@pytest.fixture
def admin_client(tesla_ce_system):
    """
        Initialize the TeSLA client.

        :return: TeSLA Client with Administration credentials
    """
    # Generate credentials for API and LAPI
    credentials_api = tesla_ce_system.vault.get_module_credentials('api')
    credentials_lapi = tesla_ce_system.vault.get_module_credentials('lapi')
    role_id = [credentials_api['role_id'], credentials_lapi['role_id']]
    secret_id = [credentials_api['secret_id'], credentials_lapi['secret_id']]

    config = ConfigManager(load_config=False)
    config.config.set('VAULT_SSL_VERIFY', tesla_ce_system.config.config.get('VAULT_SSL_VERIFY'))
    config.load_vault(vault_url=tesla_ce_system.config.config.get('VAULT_URL'),
                      role_id=role_id,
                      secret_id=secret_id,
                      approle_path=tesla_ce_system.config.config.get('VAULT_MOUNT_PATH_APPROLE'),
                      kv_path=tesla_ce_system.config.config.get('VAULT_MOUNT_PATH_KV')
                      )
    settings.TESLA_CONFIG=config
    settings.TESLA_MODULES=config.enabled_modules

    yield tesla_ce_system


@pytest.fixture
def api_client(admin_client):
    return create_module_client(admin_client, 'api')


@pytest.fixture
def lapi_client(admin_client):
    return create_module_client(admin_client, 'lapi')


@pytest.fixture
def rest_api_client(admin_client, monkeypatch):
    from rest_framework.test import APIClient
    client = APIClient()
    return client


@pytest.fixture
def user_global_admin(db):
    """
        Get a user with Global Administration privileges

        :return: User with Global Admin privileges
    """
    from tesla_ce.models import User

    admin = User.objects.filter(is_staff=True).all()
    if len(admin) == 0:
        admin = User.objects.create_user('test_admin', 'test_admin@tesla-ce.eu', is_staff=True, is_active=True)
    else:
        admin = admin[0]

    yield admin


@pytest.fixture
def institution_test_case(db):
    """
        Get a test institution with one user

        :return: Dictionary with institution and user
    """
    from tesla_ce.models import Institution, InstitutionUser

    # Create a test institution
    test_inst = Institution.objects.create(
        name="PyTest Test institution",
        acronym=get_random_string(10)
    )

    # Create a user for this institution
    user_name = get_random_string(10)
    test_user = InstitutionUser.objects.create(
        username=user_name,
        uid=user_name,
        email='{}@tesla-ce.eu'.format(user_name),
        first_name=user_name[:5],
        last_name=user_name[5:],
        institution=test_inst,
        login_allowed=False,
    )

    yield {
        'institution': test_inst,
        'user': test_user.user_ptr
    }


@pytest.fixture
def institution_course_test_case(db, admin_client, institution_test_case):
    """
        Get a test institution with one user, one course and users

        :return: Dictionary with institution and user
    """
    from tesla_ce.models import Course, InstitutionUser, Instructor, Learner, VLE

    # Create a VLE
    vle_name = get_random_string(10)
    admin_client.register_vle(
        type='moodle',
        name=vle_name,
        url='{}.tesla-ce-eu'.format(vle_name),
        institution_acronym=institution_test_case['institution'].acronym,
        client_id=get_random_string(15)
    )
    test_vle = VLE.objects.get(name=vle_name)

    # Create a test course
    test_course = Course.objects.create(
        code=get_random_string(10),
        description="PyTest test course",
        vle=test_vle,
        vle_course_id=get_random_string(5)
    )

    # Create an instructor
    instructor_user_name = get_random_string(10)
    test_instructor_usr = InstitutionUser.objects.create(
        username=instructor_user_name,
        uid=instructor_user_name,
        email='{}@tesla-ce.eu'.format(instructor_user_name),
        first_name=instructor_user_name[:5],
        last_name=instructor_user_name[5:],
        institution=institution_test_case['institution'],
        login_allowed=False,
    )
    test_instructor = Instructor(
        institutionuser_ptr=test_instructor_usr,
        institution=institution_test_case['institution']
    )
    test_instructor.save_base(raw=True)
    test_instructor = Instructor.objects.get(id=test_instructor_usr.id)
    test_course.instructors.add(test_instructor)

    # Create a learner
    learner_user_name = get_random_string(10)
    test_learner_usr = InstitutionUser.objects.create(
        username=learner_user_name,
        uid=learner_user_name,
        email='{}@tesla-ce.eu'.format(learner_user_name),
        first_name=learner_user_name[:5],
        last_name=learner_user_name[5:],
        institution=institution_test_case['institution'],
        login_allowed=False,
    )
    test_learner = Learner(
        institutionuser_ptr=test_learner_usr,
        institution=institution_test_case['institution'],
        joined_at=timezone.now()
    )
    test_learner.save_base(raw=True)
    test_learner = Learner.objects.get(id=test_learner_usr.id)
    test_course.learners.add(test_learner)

    institution_test_case['vle'] = test_vle
    institution_test_case['course'] = test_course
    institution_test_case['instructor'] = test_instructor.user_ptr
    institution_test_case['learner'] = test_learner.user_ptr

    yield institution_test_case


@pytest.fixture()
def config_mode_settings(settings):
    """
        Initialize the TeSLA client.

        :return: TeSLA Client with Administration credentials
    """
    settings.TESLA_CONFIG.config.set('TESLA_MODE', 'config')


@pytest.fixture
def empty_ui_routes(db):
    """
        Remove all UI Options routes
    """
    from tesla_ce.models import UIOption
    UIOption.objects.all().delete()

    yield []


@pytest.fixture
def base_ui_routes(db, empty_ui_routes):
    """
        Get a test institution with one user, one course and users

        :return: Dictionary with institution and user
    """
    from tesla_ce.models import UIOption
    routes = []
    # Administration route
    routes.append('/admin')
    UIOption.objects.create(route='/admin', enabled=True, roles='GLOBAL_ADMIN')
    # Dashboard route
    routes.append('/dashboard')
    UIOption.objects.create(route='/dashboard', enabled=True)
    # MyCourses route
    routes.append('/mycourses')
    UIOption.objects.create(route='/mycourses', enabled=True, roles='INSTRUCTOR,LEARNER')
    # Institution admin
    routes.append('/inst_admin')
    UIOption.objects.create(route='/inst_admin', enabled=True, roles='ADMIN')

    # Invalidate the cache
    from tesla_ce.models.ui_option import get_role_base_ui_routes
    get_role_base_ui_routes.invalidate()

    yield routes



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA decorators """
from functools import wraps

from django.conf import settings

from .exception import TeslaConfigException


def tesla_mode_required(modes):
    """
        Decorator to restrict the use of methods to a list of tesla running modes.

        :param function: Function to be decorated
        :param modes: List of modes. If not provided, method is restricted only to production
        :type modes: list
    """
    def decorator(function):
        @wraps(function)
        def _wrapped_function(*args, **kwargs):
            valid_modes = modes
            if valid_modes is None:
                valid_modes = ['production']
            if not isinstance(valid_modes, list):
                valid_modes = [valid_modes]
            current_mode = settings.TESLA_CONFIG.config.get('TESLA_MODE')
            if current_mode not in valid_modes:
                raise TeslaConfigException('Method not allowed in {} mode.'.format(current_mode))
            return function(*args, **kwargs)
        return _wrapped_function
    return decorator



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" User views module """
from django.db.models import Q
from django_filters import rest_framework as filters_dj
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import UserSerializer
from tesla_ce.models import Institution
from tesla_ce.models import User


def institution_values():
    """ Get available institutions for filter """
    institutions = Institution.objects.all()
    return ((inst.id, inst.acronym) for inst in institutions)


def roles_values():
    """ Get available roles for filter """
    return (
        ('GLOBAL_ADMIN', 'GLOBAL_ADMIN'),
        ('ADMIN', 'ADMIN'),
        ('SEND', 'SEND'),
        ('LEGAL', 'LEGAL'),
        ('DATA', 'DATA'),
        # ('INSTRUCTOR', 'INSTRUCTOR'),
        # ('LEARNER', 'LEARNER'),
    )


class UserFilter(filters_dj.FilterSet):
    """ Filter implementation for Users """
    institution = filters_dj.ChoiceFilter(field_name="institution",
                                          label="institution",
                                          choices=institution_values,
                                          null_label='None',
                                          null_value=-1,
                                          method='filter_institution')

    roles = filters_dj.MultipleChoiceFilter(field_name="roles",
                                            label="roles",
                                            choices=roles_values,
                                            method='filter_roles')

    username = filters_dj.CharFilter(lookup_expr='icontains')
    first_name = filters_dj.CharFilter(lookup_expr='icontains')
    last_name = filters_dj.CharFilter(lookup_expr='icontains')
    email = filters_dj.CharFilter(lookup_expr='icontains')

    def filter_institution(self, queryset, name, value):

        # Get the value
        value = int(value)

        # Request users with no institution assigned
        if value < 0:
            return queryset.filter(institutionuser=None)

        # Filter by provided institution
        return queryset.filter(institutionuser__institution=value)

    def filter_roles(self, queryset, name, value):

        # No filter provided
        if value is None or len(value) == 0:
            return queryset

        # Filter by provided roles
        filter_q = Q()
        if 'GLOBAL_ADMIN' in value:
            filter_q = filter_q | Q(is_staff=True)
        if 'ADMIN' in value:
            filter_q = filter_q | Q(institutionuser__inst_admin=True)
        if 'SEND' in value:
            filter_q = filter_q | Q(institutionuser__send_admin=True)
        if 'LEGAL' in value:
            filter_q = filter_q | Q(institutionuser__legal_admin=True)
        if 'DATA' in value:
            filter_q = filter_q | Q(institutionuser__data_admin=True)
        return queryset.filter(filter_q)

    class Meta:
        model = User
        fields = ['username', 'first_name', 'last_name', 'email', 'institution', 'roles']


# pylint: disable=too-many-ancestors
class AdminUserViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    model = User
    serializer_class = UserSerializer
    permission_classes = [permissions.GlobalAdminPermission]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_class = UserFilter
    search_fields = ['username', 'first_name', 'last_name', 'email']

    def get_queryset(self):
        queryset = User.objects
        return queryset.all().order_by('id')



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for instrument administration """
import logging

import pytest

import tests.utils


##############################################################################
#
#   ADMIN TESTING: INSTRUMENTS
#   1) List instruments --> 2) Read 1st instrument --> 3) Read non existing instrument
#   4) Create instrument --> 5) List instruments (+1) --> 6) Read instrument --> 7) failed create instrument
#   8) update instrument --> 9) read instrument --> 10) delete instrument --> 11) list (-1)
#
##############################################################################
@pytest.mark.django_db
def test_api_admin_instruments(rest_api_client, user_global_admin):
    # TODO: add list of not allowed users as 3rd parameter to check_403 function
    # Check that unauthenticated calls are not allowed
    user_global_admin.is_staff = False
    user_global_admin.save()
    tests.utils.check_403(rest_api_client, '/api/v2/admin/instrument/', [user_global_admin])

    # Authorization – JWT with Global Administration privileges
    # Set global admin user.
    user_global_admin.is_staff = True
    user_global_admin.save()
    rest_api_client.force_authenticate(user=user_global_admin)

    logging.info('\nTESTING ORGANIC ADMIN: INSTRUMENTS*********************************************')
    """ ---------------------------------------------------------------------
    LIST INSTRUMENTS:
       GET /api/v2/admin/instrument/
       Status Codes:
           200 OK – Ok
    """
    logging.info('\n1) LIST INSTRUMENTS --------------------------------------')
    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                           'List instruments', 'RESPONSE:', 200)

    """
    Getting values for testing process
        instruments_empty: boolean value. If True, reading specific instrument is skipped.
        n_instruments: number of instruments.
        id_first_instrument: ID from first Instrument.
        id_non_existing_instrument: taking into account IDs are assigned in ordered way.
    """
    instruments_empty = True
    id_non_existing_instrument = 1000000
    n_instruments = body['count']
    if n_instruments > 0:
        instruments_empty = False
        id_first_instrument = body['results'][0]['id']

        # pagination test
        tests.utils.check_pagination(rest_api_client, body)

        args = [['List instruments'], ['ID FIRST INSTRUMENT: ', id_first_instrument],
                ['NUMBER OF INSTRUMENTS:', n_instruments],
                ['ID NON EXISTING INSTRUMENT:', id_non_existing_instrument]]
    else:
        id_first_instrument = -1
        args = [['List instruments'], ['Empty instruments list. Reading first instrument test will be skipped.']]

    tests.utils.print_log(args)

    """ -----------------------------------------------------------------------------
     Read instrument information:
       GET /api/v2/admin/instrument/(int: instrument_id)/
       Status Codes:
           200 OK – Ok
           404 Not Found – Instrument not found
    -----------------------------------------------------------------------------"""
    # READ existing 1st instrument
    logging.info('\n2) READ 1st INSTRUMENT --------------------------------------')

    if not instruments_empty:
        str_response = 'RESPONSE id=' + str(id_first_instrument) + ':'
        tests.utils.get_rest_api_client(rest_api_client,
                                        '/api/v2/admin/instrument/' + str(id_first_instrument) + '/',
                                        'Read Instrument Information', str_response, 200)

    # READ non existing instrument with numeric id
    str_url = '/api/v2/admin/instrument/' + str(id_non_existing_instrument) + '/'
    str_response = 'RESPONSE id=' + str(id_non_existing_instrument) + ':'
    tests.utils.get_rest_api_client(rest_api_client, str_url,
                                    'Read Instrument Information', str_response, 404)

    """ -----------------------------------------------------------------------------
    Create a new instrument:
       POST /api/v2/admin/instrument/
       Status Codes:
           201 Created – Created
           400 Bad Request – Invalid information provided. The response contains the description of the errors.
    -----------------------------------------------------------------------------"""
    """
       Create instrument --> List instruments(+1) --> Read instrument --> 
       update instrument --> read instrument --> delete instrument --> list(-1)
       """

    # CREATE new instrument: SUCCESS
    logging.info('\n4) CREATE INSTRUMENT --------------------------------------')
    instrument = {'acronym': 'n3', 'name': 'name3', 'queue': 'tesla_test23'}
    new_instrument_id = tests.utils.post_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                                         instrument, 'Create New Instrument', 'RESPONSE:', 201)

    # 666 TODO:DOC paràmetres doc help no diu que options_schema i queue són OBLIGATORIS. Ho afegeixo???
    # 666 primers errors RETORNA RESPONSE:
    # {'options_schema': ['This field is required.'], 'queue': ['This field is required.']}

    # LIST instruments and check count has increased.
    logging.info('\n5) LIST INSTRUMENT --------------------------------------')
    aux_body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                               'Create New Instrument', 'RESPONSE:', 200)
    # Get number of instruments
    aux_n_instruments = aux_body['count']
    str_log = [['Create New Instrument'], ['Number of Instruments PRE:', n_instruments, 'and POST:', aux_n_instruments]]
    tests.utils.print_log(str_log)
    assert aux_n_instruments == n_instruments + 1

    # READ new instrument
    logging.info('\n6) READ NEW INSTRUMENT --------------------------------------')
    str_url = '/api/v2/admin/instrument/' + str(new_instrument_id) + '/'
    tests.utils.get_rest_api_client(rest_api_client, str_url,
                                    'Create New Instrument [Read Instrument Information]',
                                    'RESPONSE id=', 200)

    # CREATE new instrument: FAILED
    logging.info('\n7) FAILED CREATION NEW INSTRUMENT --------------------------------------')
    failed_instrument = {'acronym': 'n3'}
    tests.utils.post_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                     failed_instrument, 'Create New Instrument',
                                     'RESPONSE:', 400)

    """ -----------------------------------------------------------------------------
     Update instrument information:
       PUT /api/v2/admin/instrument/(int: instrument_id)/
       Status Codes:
           200 OK – Ok
           400 Bad Request – Invalid information provided. The response contains the description of the errors.
           404 Not Found – Instrument not found 
    -----------------------------------------------------------------------------"""

    # [Status: 200] UPDATE new instrument
    logging.info('\n8) UPDATE INSTRUMENT --------------------------------------')
    str_data = {'acronym': 'CHANGED', 'name': 'name3', 'queue': 'tesla_test23'}
    str_path = '/api/v2/admin/instrument/' + str(new_instrument_id) + '/'
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update instrument information',
                                    'RESPONSE:', 200)
    # [Status: 400] UPDATE WRONG NUMBER PARAMETERS. REQUIRED: name, queue
    str_data = {'acronym': 'CHANGED'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update instrument information',
                                    'RESPONSE:', 400)

    # [Status: 404] UPDATE NON EXISTING instrument
    str_path = '/api/v2/admin/instrument/' + str(new_instrument_id + 1) + '/'
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update instrument information',
                                    'RESPONSE:', 404)

    # Read updated instrument
    logging.info('\n9) READ UPDATED INSTRUMENT --------------------------------------')
    str_path = '/api/v2/admin/instrument/' + str(new_instrument_id) + '/'
    str_response = 'RESPONSE id=' + str(new_instrument_id) + ':'
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Updated Instrument', str_response, 200)
    """-----------------------------------------------------------------------------
     Delete an instrument:
       DELETE /api/v2/admin/instrument/(int: instrument_id)/
       Status Codes:
           204 No Content – No Content
           404 Not Found – Instrument not found
    -----------------------------------------------------------------------------"""
    logging.info('\n10) DELETE UPDATED INSTRUMENT --------------------------------------')
    # Delete updated instrument
    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                           'List instruments', 'RESPONSE:', 200)
    n_instruments_pre = body['count']
    if n_instruments > 0:
        id_first_instrument_pre = body['results'][0]['id']
        while body['next'] is not None:
            body = tests.utils.get_rest_api_client(rest_api_client, body['next'], 'List instruments', 'Pagination', 200)
        id_last_instrument_pre = body['results'][len(body['results']) - 1]['id']
    else:
        id_first_instrument_pre = -1
        id_last_instrument_pre = -1

    # [Status = 204] Delete existing item
    tests.utils.delete_rest_api_client(rest_api_client, str_path, 'Delete an instrument',
                                       'DELETED', 204)

    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                           'List instruments', 'RESPONSE:', 200)
    n_instruments_post = body['count']
    if n_instruments > 0:
        id_first_instrument_post = body['results'][0]['id']
        while body['next'] is not None:
            body = tests.utils.get_rest_api_client(rest_api_client, body['next'], 'List instruments', 'Pagination', 200)
        id_last_instrument_post = body['results'][len(body['results']) - 1]['id']
    else:
        id_first_instrument_post = -1
        id_last_instrument_post = -1

    str_log = [['Delete an instrument'],
               ['PRE delete: ', n_instruments_pre, 'items - ID first item:',
                id_first_instrument_pre, '- ID last item:', id_last_instrument_pre]]
    tests.utils.print_log(str_log)

    str_log = [['Delete an instrument'],
               ['POST delete:', n_instruments_post, 'items - ID first item:',
                id_first_instrument_post, '- ID last item:', id_last_instrument_post]]
    tests.utils.print_log(str_log)

    # [Status = 404] DELETE non existing item
    tests.utils.delete_rest_api_client(rest_api_client, str_path, 'Delete an instrument',
                                       'NOT DELETED', 404)

    logging.info('\n11) LIST INSTRUMENTS -----------------------------------------------')
    tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/instrument/',
                                    'List instruments', 'RESPONSE:', 200)

    # UPDATE new instrument FAILED
    # {"id": 12, "options_schema": null, "name": "name666", "acronym": "n666", "queue": "tesla_666",
    # "enabled": false, "requires_enrolment": false, "description": "testing", "identity": false,
    # "originality": false, "authorship": false, "integrity": false,
    # "created_at": "2021-03-21T08:48:13.866822Z", "updated_at": "2021-03-21T08:48:13.866896Z"}
